import pandas as pd
import json
from typing import Dict, List
from dsframework.base.common.functions import flatten


class ZIDS_Tester:

    def __init__(self):
        self.truth_dataset_id = -1
        self.model_type_id = -1
        self.data_keys = []
        self.pred_name = ""
        self.target_name = ""

    def save_meta_data(self, input_req):
        """
        Save input request meta data.
        Parameters:
            input_req: An object of type TestInputRequest
        """
        self.truth_dataset_id = input_req.truth_dataset_id
        self.model_type_id = input_req.model_type_id

    def test_batch(self, test_request):
        raise NotImplementedError

    def convert_input_to_dataframe(self, data_rows):
        """
        Convert from a list of TestInputRow to a dataframe.
        Also apply Model-Specific conversion.

        Parameters:
            data_rows: A list of objects, List[TestInputRow]
        Returns:
            Pandas dataframe that contains all the relevant information
        """
        df = pd.DataFrame.from_records((data_row.dict() for data_row in data_rows))
        df = df.apply(self.adapt_to_dsp_format, axis=1)
        return df

    def create_model_request(self, batch_input_rows):
        raise NotImplementedError

    def run_model_on_data_rows(self, row_dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Run model on all data rows. Create a format the the model can work with, and run the pipeline.
        Parameters:
            row_dataframe: Input dataframe, containing all data rows.
        Returns:
            Extended dataframe which contains the output of the model
        """
        batch_of_pages = row_dataframe.to_dict(orient='records')

        model_request = self.create_model_request(batch_of_pages)

        # Call the model to get predictions
        predicted_records = self.pipeline.execute(model_request)

        if predicted_records:
            df_predicted = self.create_predicted_records_dataframe(predicted_records)
            row_dataframe = row_dataframe.join(df_predicted)
        else:
            # we did not get any records, create empty records
            self.create_empty_predictions(row_dataframe)

        return row_dataframe

    def adapt_to_dsp_format(self, df_row):
        """
        Adapt for DSP format compatibility
        'data' is provided as string since the DSP saves it this way.
        Please feel free to modify due to your project needs

        Parameters:
            df_row: A single row from the processed data frame
        Returns:
            df_row: The flattened data frame, with all the fields previously compressed in the 'data' field
        """

        # Extract keys for later use
        self.data_keys = json.loads(df_row['data']).keys()

        # Restructure as a data frame line
        for k, v in json.loads(df_row['data']).items():
            df_row[k] = v
        return df_row

    def create_predicted_records_dataframe(self, predicted_records):
        # create a dataframe formatted as input for evaluation
        df_predicted = pd.DataFrame(predicted_records)
        df_predicted = df_predicted.rename(columns={self.pred_name: 'pred_model'})
        return df_predicted

    @staticmethod
    def create_empty_predictions(row_dataframe):
        """
        Create empty predictions in case where the model returns an error and we wish to keep evaluating
        """
        raise NotImplementedError

    def format_evaluation_output(self, row_dataframe) -> List[Dict]:
        """
        Make the results fit with the expected DSP data format
        :param row_dataframe:
        :return:
        """

        row_dataframe = row_dataframe.rename(columns={'pred': 'target', 'pred_model': 'pred'})

        output = row_dataframe.to_dict(orient='records')
        for rec in output:
            pred_dict = {}
            target_dict = {}
            for row_key in self.data_keys:
                pred_dict[row_key] = rec[row_key]
                target_dict[row_key] = rec[row_key]
            pred_dict['pred'] = rec['pred']
            target_dict['pred'] = rec['target']

            for row_key in self.data_keys:
                del rec[row_key]

            rec['target'] = json.dumps(pred_dict)
            rec['pred'] = json.dumps(target_dict)

        return output

    @staticmethod
    def format_response_row_for_dsp_upload(output, test_request):
        """
        Adapt response format to the DSP requirements
        :param output:
        :param test_request:
        :return:
        """
        row_extra_data = {
            "truth_dataset_id": test_request.truth_dataset_id,
            "model_type_id": test_request.model_type_id
        }
        for row in output:
            row.update(row_extra_data)
            row['truth_id'] = row.get('id', -1)

            if 'id' in row:
                del row['id']

            confusion_matrix_flattened = flatten(row['confusion_matrix'])
            row.update(confusion_matrix_flattened)

            del row['confusion_matrix']

            if 'id' in row:
                del row['data']

    @staticmethod
    def get_f1_precision_recall(var_name, df):
        """
        Calculate from detailed output report the f1, precision and recall scores per attribute.
        @param var_name: variable name to calculate the scores for.
        @param df: detailed results dataframe with per attribute confusion matrix as outputted by tester.
        @return: f1, precision and recall scores.
        This function should be overridden if other form of calculation is needed
        """
        tp = df[var_name + '_tp']
        fp = df[var_name + '_fp']
        tn = df[var_name + '_tn']
        fn = df[var_name + '_fn']
        recall = 0
        precision = 0
        f1 = 0
        if len(np.where(tp + fp + tn + fn != 1)[0]) > 0:
            print(f'{len(np.where(tp + fp + tn + fn != 1)[0])} examples have errors')
        if (tp.sum() + fn.sum()) != 0:
            recall = tp.sum() / (tp.sum() + fn.sum())
        if (tp.sum() + fp.sum()) != 0:
            precision = tp.sum() / (tp.sum() + fp.sum())
        if (precision + recall) != 0:
            f1 = 2 * ((precision * recall) / (precision + recall))
        return f1, precision, recall


#!/usr/bin/env python
# coding: utf-8

class ZIDS_Shared_data():
    artifacts_example = -1

    @staticmethod
    def get_artif():
        return ZIDS_Shared_data.artifacts_example

    @staticmethod
    def set_artif(artif):
        ZIDS_Shared_data.artifacts_example = artif






class CloudEvalClientBase:
    UNIMPLEMENTED_ERROR_MESSAGE = "Please override the parent base class methods for using cloud eval"

    def get_request_headers(self, row: dict):
        """
        construct the model service request headers.
        dataset csv record is provided in case it is needed.
        Ex:
        ```
        return {
            "accept": "application/json",
            "Content-Type": "application/json",
            "x-token": "fake-super-secret-token"
        }
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)

    def get_request_payload(self, row: dict):
        """
        construct a model service payload from a dataset csv record
        Ex:
        ```
        data = json.loads(record['data'])

        service_input = PyScoopsClassificationInputs(html_content=data['text'],
                                                     source=data['source'],
                                                     queue_name=data['queue_name'],
                                                     title=data['title']
                                                     )
        model_request = service_input  # [input]
        payload = model_request.dict()
        return payload
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)

    def extract_model_predictions_from_response(self, predictions):
        """
        Extract the predictions from the service response.
        ```
        if type(predictions) == list:
            if len(predictions) > 1:
                raise Exception("We do not currently support mini batches")
            prediction = predictions[0]
            try:
                prediction_obj = PyScoopsClassificationOutputs(**prediction)
                return prediction_obj.dict()
            except:
                # default answer
                return {}
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)

    def get_endpoint_name(self):
        """
        Return the model's endpoint name.
        For example, if you invoke the model with `/predict` then you can do:
        ```
        return "predict"
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)

    def evaluate_model_results(self, row: dict):
        """
        Evaluate results received from the model by comparing them to desired outputs
        Ex: (From scoops)
        ```
        # Load evaluator
        service_evaluator = PyScoopsClassificationEvaluator()

        # Extract specific row prediction
        row_prediction_dict = json.loads(row['prediction'])
        row_prediction = int(row_prediction_dict['pred'])

        # Extract target from row information
        data_dict = json.loads(row['data'])
        target = int(data_dict['pred'])

        # Perform evaluation
        conf_matrix_dict = service_evaluator.get_confusion_matrix(target, row_prediction)

        return conf_matrix_dict
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)

    def set_format_output(self, row: dict):
        """
        Fit specific information received from the model / evaluation to the desired output format
        Ex: (From scoops)
        ```
        # Format the dictionary according to the DSP format

        # Extract data from original dataset row
        data_dict = json.loads(row['data'])

        # DSP expects to get the exactly same data that it sent, just with the 'pred' field different
        basic_dict = {'html_content': data_dict['text'],
                      'source': data_dict['source'],
                      'queue_name': data_dict['queue_name'],
                      'title': data_dict['title'],
                      'internal_notes': data_dict['internal_notes']}

        pred_dict = {'pred': int(json.loads(row['prediction'])['pred'])}
        target_dict = {'pred': int(data_dict['pred'])}
        pred_dict_to_dsp = dict(basic_dict, **pred_dict)
        target_dict_to_dsp = dict(basic_dict, **target_dict)

        row['target'] = json.dumps(target_dict_to_dsp)
        row['pred'] = json.dumps(pred_dict_to_dsp)
        row['prob'] = json.loads(row['prediction'])['prob']

        confusion_matrix_flattened = flatten(row['evaluation'])
        row.update(confusion_matrix_flattened)

        # Remove unnecessary fields
        remove_key = row.pop('evaluation', None)
        remove_key = row.pop('prediction', None)
        remove_key = row.pop('data', None)
        ```
        """
        raise NotImplementedError(CloudEvalClientBase.UNIMPLEMENTED_ERROR_MESSAGE)
